"""
@author: Mirko Polato
@contact: mpolato@math.unipd.it
@organization: University of Padua
"""

# Base class for recommendation engine
class RecEngine(object):

	def __init__(self, data, verbose=True):
		"""
		@param data: the dataset
		@type data: Dataset
		"""
		self.data = data
		self.n_users = data.num_users()
		self.n_items = data.num_items()
		self.verbose = verbose

	# Trains the model
	def train(self, test_users=None):
		"""
		@return: the model itself
		@rtype: RecEngine
		"""
		return self
	
	# Returns the items' score for the user
	def get_scores(self, u):
		raise NotImplementedError

	# Gets the method's parameters
	def get_params(self):
		return {}
	
	# Sets the method's parameters
	def set_params(self, **params):
		"""
		@param params: parameters of the method
		@type params: dictionary {string: any_type}
		@return: the model itself
		@rtype: RecEngine
		"""
		for parameter, value in params.items():
			self.setattr(parameter, value)
		return self
	
	# Returns the name of the class with its parameters
	def get_fullname(self):
		return "%s : %s" %(self.__class__.__name__, self.get_params())




"""
@author: Mirko Polato
@contact: mpolato@math.unipd.it
@organization: University of Padua
"""

import numpy as np
import math
from pyros.utils import misc


# Calculates the AUC
def auc(ord_pred, target):
	"""
	@param ord_pred: ordere predicted scores
	@type ord_pred: iterable sequence of int
	@param target: set of positive items
	@type target: set of int
	@return: AUC
	@rtype: float
	"""
	np = len(target)
	n = len(ord_pred)
	area = 0.0
	
	for i in range(n):
		if ord_pred[i] in target:
			for j in range(i+1, n):
				if ord_pred[j] not in target:
					area += 1.0
					
	area /= float(np * (n - np))
	return area


# Calculates the average precision at k
def ap_k(ord_pred, target, k=100):
	if len(ord_pred) > k:
		ord_pred = ord_pred[:k]

	score = 0.0
	num_hits = 0.0

	for i,p in enumerate(ord_pred):
		if p in target:
			num_hits += 1.0
			score += num_hits / (i+1.0)

	if not target:
		return 1.0

	return score / min(len(target), k)


# Calculates the mean average precision at k
def map_k(ord_pred_all, target_all, k=10):
	return np.mean([ap_k(p, a, k) for p,a in zip(ord_pred_all, target_all)])


# Calculates the normalized discounted cumulative gain at k
def ndcg_k(ord_pred, target, k=100):
	k = min(k, len(ord_pred))
	idcg = idcg_k(k)
	dcg_k = sum([int(ord_pred[i] in target) / math.log(i+2, 2) for i in xrange(k)])
	return dcg_k / idcg


# Calculates the ideal discounted cumulative gain at k
def idcg_k(k):
	res = sum([1.0/math.log(i+2, 2) for i in xrange(k)])
	if not res:
		return 1.0
	else:
		return res


def evaluate(rec, ts):
	tot = {"auc":0.0, "ndcg":0.0, "ap":0.0}
	
	for u in ts.users:
		pred = misc.sort(rec.get_scores(u), rec.data.get_items(u))
		t = ts.get_items(u)
		
		area = auc(pred, t)
		ndcg = ndcg_k(pred, t)
		ap = ap_k(pred, t)
		
		tot["auc"] += area
		tot["ndcg"] += ndcg
		tot["ap"] += ap
		
		#logging.info("%d = {auc:%s, ap:%s, ndcg:%s}" %(i, area, ap, ndcg))
		
	for metric in tot:
		tot[metric] /= float(len(ts.users))

	return tot



"""
@author: Mirko Polato
@contact: mpolato@math.unipd.it
@organization: University of Padua
"""

import numpy as np
from scipy.special import binom

def fast_generalized_md_kernel(K0, n, D=2):
	N = np.full(K0.shape, n)
	XX = np.dot(np.diag(K0).reshape(K0.shape[0],1), np.ones((1,K0.shape[0])))
	N_x = N - XX
	N_xz = N_x - XX.T + K0
	N_d, N_xd, N_xzd = N.copy(), N_x.copy(), N_xz.copy()
	
	yield 1, N_d - N_xd - N_xd.T + N_xzd
	for d in range(1,D):
		N_d = N_d * (N - d) / (d + 1)
		N_xd = N_xd * (N_x - d) / (d + 1)
		N_xzd = N_xzd * (N_xz - d) / (d + 1)
		yield d+1, N_d - N_xd - N_xd.T + N_xzd
	
	
def fast_generalized_mc_kernel(K0, n, C=2):
	K = K0.copy()	
	yield 1, K
	for c in range(1,C):
		K = K * (K0 - c) / (c + 1)
		yield c+1, K


def fast_generalized_c_kernel(K0, n, C=2):
	N = np.full(K0.shape, n)
	XX = np.dot(np.diag(K0).reshape(K0.shape[0],1), np.ones((1,K0.shape[0])))
	K1 = N - XX - XX.T + 2*K0
	yield 1, K1
	K = K1.copy()
	for c in range(1,C):
		K = K * (K1 - c) / (c + 1)
		yield c+1, K


def fast_generalized_d_kernel(K0, n, D=2):
	c_gen = fast_generalized_c_kernel(K0, n, C=D)
	N = np.full(K0.shape, n)
	Z = np.full(K0.shape, 2)
	T = np.full(K0.shape, 2)
	K = c_gen.next()[1]
	yield 1, K
	for d in range(1,D):
		N = N * (n - d) / (d + 1)
		Z = Z * T
		yield d+1, (Z - T) * N + c_gen.next()[1]


def fast_generalized_mdnf_kernel(K0, n, D=2, C=2):
	for c, cK in fast_generalized_mc_kernel(K0, n, C):	
		for d, dK in fast_generalized_md_kernel(cK, binom(n, c), D):
			yield d, c, dK


def fast_generalized_mcnf_kernel(K0, n, D=2, C=2):
	for d, dK in fast_generalized_md_kernel(K0, n, D):
		for c, cK in fast_generalized_mc_kernel(dK, binom(n, d), C):
			yield d, c, cK

	
def fast_generalized_dnf_kernel(K0, n, D=2, C=2):
	for c, cK in fast_generalized_c_kernel(K0, n, C):
		for d, dK in fast_generalized_md_kernel(cK, 2**c*binom(n, c), D):
			yield d, c, dK


def fast_generalized_cnf_kernel(K0, n, D=2, C=2):
	for d, dK in fast_generalized_d_kernel(K0, n, D):
		for c, cK in fast_generalized_mc_kernel(dK, 2**d*binom(n, d), C):
			yield d, c, cK


def mc_kernel(X, c=1):
	return [k for _,k in fast_generalized_mc_kernel(np.dot(X,X.T), X.shape[1], c)][-1]

def md_kernel(X, d=1):
	return [k for _,k in fast_generalized_md_kernel(np.dot(X,X.T), X.shape[1], d)][-1]
	
def c_kernel(X, c=1):
	return [k for _,k in fast_generalized_c_kernel(np.dot(X,X.T), X.shape[1], c)][-1]

def d_kernel(X, d=1):
	return [k for _,k in fast_generalized_d_kernel(np.dot(X,X.T), X.shape[1], d)][-1]

def mdnf_kernel(X, d=1, c=1):
	return [k for _,_,k in fast_generalized_mdnf_kernel(np.dot(X,X.T), X.shape[1], d, c)][-1]

def mcnf_kernel(X, d=1, c=1):
	return [k for _,_,k in fast_generalized_mcnf_kernel(np.dot(X,X.T), X.shape[1], d, c)][-1]

def cnf_kernel(X, d=1, c=1):
	return [k for _,_,k in fast_generalized_cnf_kernel(np.dot(X,X.T), X.shape[1], d, c)][-1]

def dnf_kernel(X, d=1, c=1):
	return [k for _,_,k in fast_generalized_dnf_kernel(np.dot(X,X.T), X.shape[1], d, c)][-1]
	
def tanimoto_kernel(X):
	d = np.array([np.dot(X[i,:],X[i,:].T) for i in xrange(X.shape[0])]).reshape((X.shape[0],1))
	Xp = np.dot(np.ones((X.shape[0],1)), d.T)
	Kl = np.dot(X, X.T)
	K = np.divide(Kl, Xp + Xp.T - Kl)
	return K

def old_dnf_kernel(X):
	return 2**np.dot(X, X.T) - 1.


"""
@author: Mirko Polato
@contact: mpolato@math.unipd.it
@organization: University of Padua
"""

import sys

from pyros.data.reader import CSVReader
import pyros.core.engine as exp
import pyros.core.evaluation as ev
import pyros.data.dataset as ds
from pyros.data.mapping import Mapping
import pyros.utils as ut
import cvxopt as co
from pyros.utils.bool_kernels import *
import numpy as np


def main_simple(argv):
	path_tr, path_ts = argv[0], argv[1]
	separator = " "
	
	# LOAD DATASET
	reader = CSVReader(path_tr, separator)
	train_set = ds.UDataset(Mapping(), Mapping())
	reader.read(train_set, True)
	#

	# LOAD TESTSET
	rdte = CSVReader(path_ts, separator)
	test_set = ds.UDataset(train_set.user_mapping, train_set.item_mapping)
	rdte.read(test_set, True)
	#
	
	#################################################
	# Uncomment *only one* of the 'rec' declaration	#
	# to test the corresponding algorithm			#
	#################################################
	
	#rec = exp.I2I_Asym_Cos(train_set)
	#rec = exp.CF_OMD(train_set)
	#rec = exp.ECF_OMD(train_set)
	#rec = exp.WRMF(train_set)
	#rec = exp.SLIM(train_set)
	#rec = exp.BPRMF(train_set)
	
	X = np.array(train_set.to_cvxopt_matrix())
	d = 4
	K = mc_kernel(X.T, d)
	rec = exp.CF_KOMD(train_set, ut.kernels.normalize(co.matrix(K)))
	
	#################################################
	
	print "Training..."
	rec.train(test_set.users)
	
	print "Evaluation..."
	result = ev.evaluate(rec, test_set)
	
	print "Done!"
	print result
	
	
#MAIN
if __name__=='__main__':
	args = sys.argv[1:]
	for i, arg in enumerate(args):
		print (i+1), arg
	
	main_simple(sys.argv[1:])


"""
@author: Mirko Polato
@contact: mpolato@math.unipd.it
@organization: University of Padua
"""

from collections import defaultdict


# Represents the identity mapping for the IDs
class IdentityMapping(object):
	
	# Adds a new explicit ID
	def add(self, explicit):
		return explicit

	# Returns the corresponding explicit ID
	def get_explicit(self, implicit):
		return implicit

	# Returns the corresponding implicit ID
	def get_implicit(self, explicit):
		return explicit


# Represents a natural mapping for the IDs
class Mapping(IdentityMapping):

	def __init__(self):
		self.map_i2e = {}
		self.map_e2i = {}
		self.map_i2e = defaultdict(lambda: -1, self.map_i2e)
		self.map_e2i = defaultdict(lambda: -1, self.map_e2i)
		self.size = 0
		
	def from_i2e(self, i2e):
		self.map_i2e = defaultdict(lambda: -1, i2e)
		self.map_e2i = defaultdict(lambda: -1, {v: k for k, v in i2e.iteritems()})
		return self
	
	def from_e2i(self, e2i):
		self.map_e2i = defaultdict(lambda: -1, e2i)
		self.map_i2e = defaultdict(lambda: -1, {v: k for k, v in e2i.iteritems()})
		return self

	def add(self, explicit):
		i = self.map_e2i[explicit]
		
		if i < 0:
			i = self.size
			self.map_e2i[explicit] = i
			self.map_i2e[i] = explicit
			self.size += 1
		
		return i

	def get_explicit(self, implicit):
		e = self.map_i2e[implicit]
		
		if e != -1: return e
		else: raise IndexError("Index %d does not exists." %implicit)

	def get_implicit(self, explicit):
		i = self.map_e2i[explicit]
		
		if i >= 0: return i
		else: raise IndexError("Index %s does not exists." %explicit)




"""
@author: Mirko Polato
@contact: mpolato@math.unipd.it
@organization: University of Padua

This module contains utility methods for manipulating numpy objects.
"""

from math import sqrt
from sklearn.preprocessing import normalize


# Normalizes the vector
def normalize_vec(v):
	"""
	@param v: the vector
	@type v: cvxopt dense matrix
	@return: the normalized vector
	@rtype: cvxopt dense matrix
	"""
	return v / sqrt(sum(v**2))


# Normalizes the rows of the matrix
def normalize(X):
	"""
	@param X: the matrix
	@type X: numpy dense matrix
	@return: the row-normalized matrix
	@rtype: numpy dense matrix
	"""
	return normalize(X)


# Normalizes the cols of the matrix
def normalize_cols(X):
	"""
	@param X: the matrix
	@type X: numpy dense matrix
	@return: the col-normalized matrix
	@rtype: numpy dense matrix
	"""
	return normalize(X, axis=0)	


# Sorts the vector by value
def sort(v, skip, dec=True):
	"""
	@param v: the vector to sort
	@type v: numpy dense matrix
	@param skip: vector of indexes to skip
	@type skip: numpy dense matrix
	@param dec: whether to sort in descending order or not
	@type dec: boolean (default: True)
	@return: ordered list by value
	@rtype: list
	"""
	d = {k:v[k] for k in xrange(v.size[0]) if k not in skip}
	return sorted(d.keys(),key=lambda s:d[s],reverse=dec)




"""
@author: Mirko Polato
@contact: mpolato@math.unipd.it
@organization: University of Padua
"""

import numpy as np

def scaled_md_kernel(K0, n, D=2):
	N = np.full(K0.shape, n)
	KXX = np.dot(np.diag(K0).reshape(K0.shape[0],1), np.ones((1,K0.shape[0])))
	XX = np.full(K0.shape, 1.)
	ZZ = XX.copy()
	XZ = XX.copy()

	for d in range(1,D+1):
		XX = np.multiply(XX, np.divide(N-KXX-d+1, N-d+1))
		ZZ = np.multiply(ZZ, np.divide(N-KXX.T-d+1, N-d+1))
		XZ = np.multiply(XZ, np.divide(N-KXX-KXX.T+K0-d+1, N-d+1))
		yield d, 1. - XX - ZZ + XZ

def scaled_mc_kernel(K0, n, C=2):
	N = np.full(K0.shape, n)
	XZ = np.full(K0.shape, 1.)

	for c in range(1,C+1):
		XZ = np.multiply(XZ, np.divide(K0-c+1, N-c+1.))
		yield c, XZ

def scaled_md(X, d):
	K0 = np.dot(X, X.T) 
	return [k for _,k in scaled_md_kernel(K0, X.shape[1], d)][-1]

def scaled_mc(X, c):
	K0 = np.dot(X, X.T) 
	return [k for _,k in scaled_mc_kernel(K0, X.shape[1], c)][-1]



"""
@author: Mirko Polato
@contact: mpolato@math.unipd.it
@organization: University of Padua
"""

import numpy as np
from sklearn.preprocessing import Binarizer as Skbin, OneHotEncoder, LabelEncoder, MinMaxScaler
from misc import save_as_svmlight
from sklearn.datasets import load_svmlight_file, dump_svmlight_file
import sys

'''
Class utility that binarize non-binary datasets
'''
class Binarizer():
	def __init__(self, M, class_index=0):
		self.M = M
		self.class_index = class_index
	
	
	def by_average(self, tune=0.0):
		tune = min(max(-1.0, tune), 1.0) # -1.0 <= tune <= 1.0
		
		A = np.mean(self.M, axis=0)
		R = np.array(self.M, copy=True)
		D = A
		if tune < 0:
			Mn = np.min(self.M, axis=0)
			D = Mn + (1. + tune)*(A - Mn)
		else:
			Mx = np.max(self.M, axis=0)
			D = A + tune*(Mx - A)
		
		for i in range(R.shape[0]):
			for j in range(R.shape[1]):
				R[i,j] = 1.0 if R[i,j] >= D[j] else 0.0
				
		return R
	
	
	def by_percentage(self, perc=.5):
		if perc <= 0. or perc > 1.: #0 <= perc <= 1
			perc = .5
		
		Mn = np.min(self.M, axis=0)
		Mx = np.max(self.M, axis=0)
		D = Mn + (Mx - Mn)*perc
		
		R = np.array(self.M, copy=True)
		for i in range(R.shape[0]):
			for j in range(R.shape[1]):
				R[i,j] = 1.0 if R[i,j] >= D[j] else 0.0
		
		return R
		
		
	def by_threshold(self, threshold=0.0):
		bin = Skbin(threshold).fit(self.M)
		return bin.transform(self.M)
	
	#TODO
	def by_entropy(self):
		pass
	
	
	# SIDE EFFECTS
	def apply_01_scaling(self):
		scaler = MinMaxScaler()
		self.M = scaler.fit_transform(self.M)
		return self
	
	
	# SIDE EFFECTS
	def apply_onehot(self, columns=[]):
		enc = OneHotEncoder()
		enc.fit(self.M[:, columns])
		R = enc.transform(self.M[:, columns]).toarray()
		self.M = np.c_[self.M[:,[x for x in range(self.M.shape[1]) if x not in columns]], R]
		self.class_index -= len([c for c in columns if c < self.class_index])
		return self
		
	
	# SIDE EFFECTS
	def to_numeric(self, columns=[]):
		le = LabelEncoder()
		for i, c in enumerate(columns):
			le.fit(self.M[:, c])
			self.M[:, c] = le.transform(self.M[:, c])
		self.M = self.M.astype(np.float)
		return self
	
	
	# SIDE EFFECTS
	def remove_columns(self, columns=[]):
		self.M = self.M[:, [x for x in range(self.M.shape[1]) if x not in columns or x == self.class_index]]
		self.class_index -= len([c for c in columns if c < self.class_index])
		return self



"""
@author: Mirko Polato
@contact: mpolato@math.unipd.it
@organization: University of Padua
"""

import cvxopt as co

def KLIN(X):
	return [X*X.T, X.size[1]]

def KXX(KA):
	n = KA.size[0]
	d = co.matrix([KA[i,i] for i in range(n)])
	U = co.matrix([1.0]*n)
	return (d * U.T)

def KNOT(K, n):
	dim = K.size[0]
	Q = co.matrix([[K[i,i]+K[j,j] for i in range(dim)] for j in range(dim)])
	return [n + K - Q, n]
	
def KAND(KA, na, KB, nb):
	n = KA.size[0]
	return [co.mul(KA,KB), na*nb]

def KOR(KA, na, KB, nb):
	return [na*nb \
			- co.mul(KNOT(KXX(KA),na)[0], KNOT(KXX(KB),nb)[0]) \
			- co.mul(KNOT(KXX(KA),na)[0], KNOT(KXX(KB),nb)[0]).T \
			+ co.mul(KNOT(KA,na)[0], KNOT(KB,nb)[0]) \
			, na * nb]

def KXOR(KA, na, KB, nb):
	return [2*KAND(KA,na,KB,nb)[0] \
			+ co.mul(KA,KNOT(KB,nb)[0]) \
			+ co.mul(KNOT(KA,na)[0],KB) \
			- co.mul(KA,KXX(KB).T) \
			+ co.mul(KXX(KA),KXX(KB).T) \
			- co.mul(KXX(KA),KB) \
			- co.mul(KA,KXX(KB)) \
			+ co.mul(KXX(KA).T,KXX(KB)) \
			- co.mul(KXX(KA).T,KB) \
			, na*nb]

def KIMP(KA, na, KB, nb):
	return KOR(KNOT(KA,na)[0],na,KB,nb)
	
def KBIMP(KA, na, KB, nb):
	return KOR(KA,na,KNOT(KB,nb)[0],nb)

def KNIMP(KA, na, KB, nb):
	return KNOT(*(KOR(KNOT(KA,na)[0],na,KB,nb)))

def KEQ(KA,na,KB,nb):
	return KNOT(*(KXOR(KA,na,KB,nb)))

def KNOR(KA,na,KB,nb):
	return KNOT(*(KOR(KA,na,KB,nb)))

def KNAND(KA,na,KB,nb):
	return KNOT(*(KAND(KA,na,KB,nb)))

def KNBIMP(KA, na, KB, nb):
	return KNOT(*(KOR(KA,na,KNOT(KB,nb)[0],nb)))



from setuptools import setup, find_packages
#from distutils.core import setup

setup(
  name = 'mkpyros',
  packages = find_packages(exclude=['build', '_docs', 'templates']),
  version = '0.9.30',
  install_requires=[
        "numpy",
        "scipy",
        "cvxopt"
  ],
  license = "MIT",
  description = 'Python module for building and evaluating recommender systems for implicit feedback. Full documentation @ http://mkpyros.readthedocs.io/en/latest/',
  author = 'Mirko Polato',
  author_email = 'mak1788@gmail.com',
  url = 'https://github.com/makgyver/pyros',
  download_url = 'https://github.com/makgyver/pyros',
  keywords = ['recommendation-algorithm', 'collaborative-filtering', 'algorithm', 'kernel'],
  classifiers = [
                 'Development Status :: 3 - Alpha',
                 'Programming Language :: Python :: 2.7',
                 'Topic :: Scientific/Engineering :: Artificial Intelligence',
                 'License :: OSI Approved :: MIT License',
                ]
)